# Copyright 2024 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

URL = {
    "protenix_base_default_v1.0.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_base_default_v1.0.0.pt",
    "protenix_base_20250630_v1.0.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_base_20250630_v1.0.0.pt",
    "protenix_base_default_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_base_default_v0.5.0.pt",
    "protenix_base_constraint_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_base_constraint_v0.5.0.pt",
    "protenix_mini_default_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_mini_default_v0.5.0.pt",
    "protenix_tiny_default_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_tiny_default_v0.5.0.pt",
    "protenix_mini_esm_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_mini_esm_v0.5.0.pt",
    "esm2_t36_3B_UR50D": "https://protenix.tos-cn-beijing.volces.com/checkpoint/esm2_t36_3B_UR50D.pt",
    "esm2_t36_3B_UR50D-contact-regression": "https://protenix.tos-cn-beijing.volces.com/checkpoint/esm2_t36_3B_UR50D-contact-regression.pt",
    "protenix_mini_ism_v0.5.0": "https://protenix.tos-cn-beijing.volces.com/checkpoint/protenix_mini_ism_v0.5.0.pt",
    "esm2_t36_3B_UR50D_ism": "https://protenix.tos-cn-beijing.volces.com/checkpoint/esm2_t36_3B_UR50D_ism.pt",
    "esm2_t36_3B_UR50D_ism-contact-regression": "https://protenix.tos-cn-beijing.volces.com/checkpoint/esm2_t36_3B_UR50D_ism-contact-regression.pt",
    "ccd_components_file": "https://protenix.tos-cn-beijing.volces.com/common/components.cif",
    "ccd_components_rdkit_mol_file": "https://protenix.tos-cn-beijing.volces.com/common/components.cif.rdkit_mol.pkl",
    "pdb_cluster_file": "https://protenix.tos-cn-beijing.volces.com/common/clusters-by-entity-40.txt",
    "obsolete_release_data_csv": "https://protenix.tos-cn-beijing.volces.com/common/obsolete_release_date.csv",
    # the following files will be used if enable_template is True
    "obsolete_pdbs_path": "https://protenix.tos-cn-beijing.volces.com/common/obsolete_to_successor.json",
    "release_dates_path": "https://protenix.tos-cn-beijing.volces.com/common/release_date_cache.json",
}
